/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2009 The LS-Universe Team

  Homepage: http://www.ls-universe.info

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "MultiMonitor.h"
#include "SettingsHelpers.h"

MultiMonitor multimonitor;

BOOL CALLBACK EnumMonitorsProc(HMONITOR hMonitor, HDC hdc, LPRECT lpRMonitor, LPARAM dwData)
{
    MONITORINFO mi = { sizeof(MONITORINFO) };
    GetMonitorInfo(hMonitor, &mi);

    MonitorEnumData *monitordata = (MonitorEnumData *) dwData;

    monitordata->pMultiMonitor->MonitorInfoArray[monitordata->pMultiMonitor->iMonitorIndex] = mi;

    RECT relativeRECT = {0,0,0,0};
    CopyRect(&relativeRECT, &mi.rcMonitor);

    monitordata->pMultiMonitor->MonitorSizeArray[monitordata->pMultiMonitor->iMonitorIndex].cx = (relativeRECT.right - relativeRECT.left);
    monitordata->pMultiMonitor->MonitorSizeArray[monitordata->pMultiMonitor->iMonitorIndex].cy = (relativeRECT.bottom - relativeRECT.top);

    monitordata->pMultiMonitor->iMonitorIndex++;

    UnionRect(monitordata->pRect, monitordata->pRect, lpRMonitor);

    if ( monitordata->pMultiMonitor->pMonitor == NULL && mi.dwFlags == MONITORINFOF_PRIMARY )
        monitordata->pMultiMonitor->pMonitor = hMonitor;
    else if ( monitordata->pMultiMonitor->sMonitor == NULL && mi.dwFlags != MONITORINFOF_PRIMARY )
        monitordata->pMultiMonitor->sMonitor = hMonitor;
    else if ( monitordata->pMultiMonitor->tMonitor == NULL && mi.dwFlags != MONITORINFOF_PRIMARY )
        monitordata->pMultiMonitor->tMonitor = hMonitor;

    return true;
}

MultiMonitor::MultiMonitor() :
    iMinScreenX(0),
    iMinScreenY(0),
    iMaxScreenWidth(0),
    iMaxScreenHeight(0),

    iMonitorIndex(0),
    pMonitor(NULL),
    sMonitor(NULL),
    tMonitor(NULL),

    hdcDesktop(NULL),
    hbmDesktop(NULL)
{
    iRealMonitors = GetSystemMetrics(SM_CMONITORS);
    InitializeMultiMonitor();   
};

MultiMonitor::~MultiMonitor()
{
    if ( hbmDesktop && hdcDesktop )
    {
        hbmDesktop = (HBITMAP) SelectObject(hdcDesktop, hbmDesktop);
        DeleteDC(hdcDesktop);
        DeleteObject(hbmDesktop);

        hbmDesktop = NULL;
        hdcDesktop = NULL;
    }
}

void MultiMonitor::GetMonitorInfo()
{
    RECT rct = {0,0,0,0};
    MonitorEnumData *monitordata = new MonitorEnumData();

    monitordata->pRect = &rct;
    monitordata->pMultiMonitor = this;

    EnumDisplayMonitors(NULL, NULL, EnumMonitorsProc, (LPARAM)monitordata);

    iMaxScreenWidth = rct.right-rct.left;
    iMaxScreenHeight = rct.bottom-rct.top;

    delete monitordata;
    monitordata = NULL;
}

void MultiMonitor::InitializeMultiMonitor()
{
    if ( !IsOS( OS_95 ) && !IsOS( OS_NT4 ) )
    {
        GetMonitorInfo();
    
        if ( iRealMonitors > 1 )
        {
            iMinScreenX = GetSystemMetrics(SM_XVIRTUALSCREEN);
            iMinScreenY = GetSystemMetrics(SM_YVIRTUALSCREEN);
        }
        else
        {
            iMinScreenX = 0;
            iMinScreenY = 0;
            iMaxScreenWidth = GetSystemMetrics(SM_CXSCREEN);
            iMaxScreenHeight = GetSystemMetrics(SM_CYSCREEN);
        }
    }
    else
    {
        iMinScreenX = 0;
        iMinScreenY = 0;
        iMaxScreenWidth = GetSystemMetrics(SM_CXSCREEN);
        iMaxScreenHeight = GetSystemMetrics(SM_CYSCREEN);
    }
}

HMONITOR GetMonitor(int iMonitor)
{
    if ( iMonitor == 1 )
        return multimonitor.pMonitor;
    else if ( iMonitor == 2 )
        return multimonitor.sMonitor;
    else if ( iMonitor == 3 )
        return multimonitor.tMonitor;
    else 
        return multimonitor.pMonitor;
}

// retrieve a string value from the registry
LPTSTR RegQueryStringValue(HKEY hKey, LPCTSTR pszSubKey, LPCTSTR pszValueName, LPTSTR pszBuffer, int nBufferSize, LPTSTR pszDefaultValue)
{
    DWORD dwDataType;
    DWORD dwDataSize = (DWORD) nBufferSize;
    HKEY hSubKey = NULL;
    LONG lErrorCode;
    
    if(pszSubKey)
    {
        lErrorCode = RegOpenKeyEx(hKey, pszSubKey, 0, KEY_QUERY_VALUE, &hSubKey);
        
        if(lErrorCode != ERROR_SUCCESS)
        {
            StringCchCopy(pszBuffer, nBufferSize, pszDefaultValue);
            return pszBuffer;
        }
        
        hKey = hSubKey;
    }
    
    lErrorCode = RegQueryValueEx(hKey, pszValueName, NULL, &dwDataType,
        (LPBYTE) pszBuffer, &dwDataSize);
    
    if(hSubKey)
        RegCloseKey(hSubKey);
        
    if(dwDataType != REG_SZ || lErrorCode != ERROR_SUCCESS)
        StringCchCopy(pszBuffer, nBufferSize, pszDefaultValue);
    
    return pszBuffer;
}

// enhanced version of PaintDesktop
void PaintDesktopEx(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, int xSrc, int ySrc, BOOL updateCache)
{
    if(multimonitor.hdcDesktop == NULL || updateCache)
    {
        if(multimonitor.hdcDesktop == NULL && hdcDest == NULL)
            return;

        if(multimonitor.hbmDesktop != NULL)
        {
            multimonitor.hbmDesktop = (HBITMAP) SelectObject(multimonitor.hdcDesktop, multimonitor.hbmDesktop);
            DeleteObject(multimonitor.hbmDesktop);
            multimonitor.hbmDesktop = NULL;
        }

        HDC hdcScreen = GetDC(NULL);

        if(multimonitor.hdcDesktop == NULL)
        {
            multimonitor.hdcDesktop = CreateCompatibleDC(hdcScreen);
        }
        
        int cxDesktop = GetSystemMetrics(SM_CXVIRTUALSCREEN);
        int cyDesktop = GetSystemMetrics(SM_CYVIRTUALSCREEN);

        multimonitor.hbmDesktop = CreateCompatibleBitmap(hdcScreen, cxDesktop, cyDesktop);
        multimonitor.hbmDesktop = (HBITMAP) SelectObject(multimonitor.hdcDesktop, multimonitor.hbmDesktop);

        RECT r;
        SetRect(&r, 0, 0, cxDesktop, cyDesktop);
        FillRect(multimonitor.hdcDesktop, &r, GetSysColorBrush(COLOR_DESKTOP));

        char wallpaper[MAX_LINE_LENGTH] = { 0 };

        RegQueryStringValue(HKEY_CURRENT_USER,
            "Control Panel\\Desktop", "Wallpaper", wallpaper,
            MAX_LINE_LENGTH, "");

        if(wallpaper[0])
        {
            int wallpaperStyle;
            int tileWallpaper;

            char buffer[32] = {0};

            RegQueryStringValue(HKEY_CURRENT_USER, "Control Panel\\Desktop", "WallpaperStyle", buffer, 32, "0");
            wallpaperStyle = atoi(buffer);

            RegQueryStringValue(HKEY_CURRENT_USER, "Control Panel\\Desktop", "TileWallpaper", buffer, 32, "0");
            tileWallpaper = atoi(buffer);

            HBITMAP hbmWallpaper = (HBITMAP) LoadImage(0, wallpaper,
                IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);

            if(hbmWallpaper)
            {
                BITMAP bm;
                GetObject(hbmWallpaper, sizeof(BITMAP), &bm);

                int cxWallpaper = bm.bmWidth;
                int cyWallpaper = bm.bmHeight;

                HDC hdcWallpaper = CreateCompatibleDC(hdcScreen);
                hbmWallpaper = (HBITMAP) SelectObject(hdcWallpaper, hbmWallpaper);

                if(tileWallpaper)
                {
                    TileBlt(multimonitor.hdcDesktop,
                        0, 0,
                        cxDesktop, cyDesktop,
                        hdcWallpaper,
                        0, 0,
                        cxWallpaper, cyWallpaper,
                        SRCCOPY);
                }
                else
                {
                    if(wallpaperStyle == 2) // stretch
                    {
                        int iUsedX = 0;
                        int iUsedY = 0;
                        int iUsedWidth = cxDesktop;
                        int iUsedHeight = cyDesktop;
                        
                        if ( multimonitor.iRealMonitors > 1 )
                        {
                            MONITORINFO mi = { sizeof(MONITORINFO) };
                            GetMonitorInfo(GetMonitor(1), &mi);

                            iUsedX = mi.rcMonitor.left-multimonitor.iMinScreenX;
                            iUsedY = mi.rcMonitor.top-multimonitor.iMinScreenY;
                            iUsedWidth = mi.rcMonitor.right-mi.rcMonitor.left;
                            iUsedHeight = mi.rcMonitor.bottom-mi.rcMonitor.top;
                        }

                        SetStretchBltMode(multimonitor.hdcDesktop, STRETCH_DELETESCANS);

                        StretchBlt(multimonitor.hdcDesktop,
                            iUsedX, iUsedY,
                            iUsedWidth, iUsedHeight,
                            hdcWallpaper,
                            0, 0,
                            cxWallpaper, cyWallpaper,
                            SRCCOPY);

                        if ( multimonitor.iRealMonitors > 1 )
                        {
                            MONITORINFO mi = { sizeof(MONITORINFO) };
                            GetMonitorInfo(GetMonitor(2), &mi);

                            iUsedX = mi.rcMonitor.left-multimonitor.iMinScreenX;
                            iUsedY = mi.rcMonitor.top-multimonitor.iMinScreenY;
                            iUsedWidth = mi.rcMonitor.right-mi.rcMonitor.left;
                            iUsedHeight = mi.rcMonitor.bottom-mi.rcMonitor.top;
                    
                            SetStretchBltMode(multimonitor.hdcDesktop, STRETCH_DELETESCANS);

                            StretchBlt(multimonitor.hdcDesktop,
                                iUsedX, iUsedY,
                                iUsedWidth, iUsedHeight,
                                hdcWallpaper,
                                0, 0,
                                cxWallpaper, cyWallpaper,
                                SRCCOPY);
                        }
                    }
                    else
                    {
                        int iUsedX = 0;
                        int iUsedY = 0;
                        int iUsedWidth = cxDesktop;
                        int iUsedHeight = cyDesktop;

                        if ( multimonitor.iRealMonitors > 1 )
                        {
                            MONITORINFO mi = { sizeof(MONITORINFO) };
                            GetMonitorInfo(GetMonitor(1), &mi);

                            iUsedX = mi.rcMonitor.left-multimonitor.iMinScreenX;
                            iUsedY = mi.rcMonitor.top-multimonitor.iMinScreenY;
                            iUsedWidth = mi.rcMonitor.right-mi.rcMonitor.left;
                            iUsedHeight = mi.rcMonitor.bottom-mi.rcMonitor.top;
                        }

                        BitBlt(multimonitor.hdcDesktop,
                                iUsedX + (iUsedWidth - cxWallpaper) / 2,
                                iUsedY + (iUsedHeight - cyWallpaper) / 2,
                                cxWallpaper, cyWallpaper,
                                hdcWallpaper,
                                0, 0,
                                SRCCOPY);

                        if ( multimonitor.iRealMonitors > 1 )
                        {
                            MONITORINFO mi = { sizeof(MONITORINFO) };
                            GetMonitorInfo(GetMonitor(2), &mi);
                            
                            iUsedX = mi.rcMonitor.left-multimonitor.iMinScreenX;
                            iUsedY = mi.rcMonitor.top-multimonitor.iMinScreenY;
                            iUsedWidth = mi.rcMonitor.right-mi.rcMonitor.left;
                            iUsedHeight = mi.rcMonitor.bottom-mi.rcMonitor.top;
                    
                            BitBlt(multimonitor.hdcDesktop,
                                iUsedX + (iUsedWidth - cxWallpaper) / 2,
                                iUsedY + (iUsedHeight - cyWallpaper) / 2,
                                cxWallpaper, cyWallpaper,
                                hdcWallpaper,
                                0, 0,
                                SRCCOPY);
                        }
                    }
                }

                hbmWallpaper = (HBITMAP) SelectObject(hdcWallpaper, hbmWallpaper);
                DeleteDC(hdcWallpaper);
                DeleteObject(hbmWallpaper);
            }
        }

        ReleaseDC(0, hdcScreen);
    }

    if(hdcDest)
    {
        if ( multimonitor.iRealMonitors > 1 )
        {
            xSrc = xSrc-multimonitor.iMinScreenX;
            ySrc = ySrc-multimonitor.iMinScreenY;
        }
        BitBlt(hdcDest, xDest, yDest, cxDest, cyDest, multimonitor.hdcDesktop, xSrc, ySrc, SRCCOPY);
    }
}