/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2009 The LS-Universe Team

  Homepage: http://www.ls-universe.info

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "PaintThumbnail.h"
#include "SettingsHelpers.h"

//DWM Thumbnails
#define DWM_TNP_RECTDESTINATION 0x00000001
//    Indicates a value for rcDestination has been specified.
#define DWM_TNP_RECTSOURCE 0x00000002
//    Indicates a value for rcSource has been specified.
#define DWM_TNP_OPACITY 0x00000004
//    Indicates a value for opacity has been specified.
#define DWM_TNP_VISIBLE 0x00000008
//    Indicates a value for fVisible has been specified.
#define DWM_TNP_SOURCECLIENTAREAONLY 0x00000010
//    Indicates a value for fSourceClientAreaOnly has been specified.

struct DWM_THUMBNAIL_PROPERTIES
{
    DWORD dwFlags;
    RECT rcDestination;
    RECT rcSource;
    BYTE opacity;
    BOOL fVisible;
    BOOL fSourceClientAreaOnly;
};

typedef HANDLE HTHUMBNAIL;
typedef HTHUMBNAIL* PHTHUMBNAIL; 

typedef HRESULT (__stdcall *PFNDwmIsCompositionEnabled)(BOOL*);

typedef HRESULT (__stdcall *PFNDwmRegisterThumbnail)(HWND, HWND*, PHTHUMBNAIL);
typedef HRESULT (__stdcall *PFNDwmUnregisterThumbnail)(HTHUMBNAIL);
typedef HRESULT (__stdcall *PFNDwmQueryThumbnailSourceSize)(HTHUMBNAIL, PSIZE);
typedef HRESULT (__stdcall *PFNDwmUpdateThumbnailProperties)(HTHUMBNAIL , const DWM_THUMBNAIL_PROPERTIES *);

static PFNDwmIsCompositionEnabled zDwmIsCompositionEnabled = NULL;

static PFNDwmRegisterThumbnail zDwmRegisterThumbnail = NULL;
static PFNDwmUnregisterThumbnail zDwmUnregisterThumbnail = NULL;
static PFNDwmQueryThumbnailSourceSize zDwmQueryThumbnailSourceSize = NULL;
static PFNDwmUpdateThumbnailProperties zDwmUpdateThumbnailProperties= NULL;

static HMODULE hModDWMAPI = NULL;

bool InitDWM(bool bCheck = false)
{
	if ( !hModDWMAPI )
	{
		HMODULE hModDWMAPI = LoadLibrary("DWMAPI.DLL");

		if ( hModDWMAPI )
		{
			zDwmIsCompositionEnabled = (PFNDwmIsCompositionEnabled)GetProcAddress(hModDWMAPI, "DwmIsCompositionEnabled");
       
			zDwmRegisterThumbnail = (PFNDwmRegisterThumbnail)GetProcAddress(hModDWMAPI, "DwmRegisterThumbnail");
			zDwmUnregisterThumbnail = (PFNDwmUnregisterThumbnail)GetProcAddress(hModDWMAPI, "DwmUnregisterThumbnail");
			zDwmQueryThumbnailSourceSize = (PFNDwmQueryThumbnailSourceSize)GetProcAddress(hModDWMAPI, "DwmQueryThumbnailSourceSize");
			zDwmUpdateThumbnailProperties = (PFNDwmUpdateThumbnailProperties)GetProcAddress(hModDWMAPI, "DwmUpdateThumbnailProperties");

			if ( zDwmIsCompositionEnabled && zDwmRegisterThumbnail && zDwmUnregisterThumbnail && zDwmQueryThumbnailSourceSize && zDwmUpdateThumbnailProperties )
			{
				BOOL bResult = FALSE;
				if ( S_OK != zDwmIsCompositionEnabled(&bResult) )
					MessageBox(NULL, "DwmIsCompositionEnabled failed!", "Info", MB_OK);
				
				return bResult;            
			}
			else
			{
				FreeLibrary(hModDWMAPI);
				hModDWMAPI = NULL;

				return false;   
			}
		}
		else
		{
			return false;
		}
	}
	else if ( bCheck )
	{
		if ( zDwmIsCompositionEnabled && zDwmRegisterThumbnail && zDwmUnregisterThumbnail && zDwmQueryThumbnailSourceSize && zDwmUpdateThumbnailProperties )
		{
			BOOL bResult = FALSE;
			zDwmIsCompositionEnabled(&bResult);
				
			return bResult;            
		}
	}

    return false;   
}

PaintThumbnail::PaintThumbnail()
{

}

PaintThumbnail::~PaintThumbnail()
{

}

BOOL PaintThumbnail::configure(LPCSTR pszName, LPCSTR pszPrefix, const PaintThumbnail *pDefaultPaintThumbnail, const PaintThumbnailDefaults *pPaintThumbnailDefaults)
{
    string name(pszName);
    string prefix(pszPrefix);

    string strAddToGroup = iGetxRCString(name, (prefix + "ThumbnailAddToGroup"), "");

    if (!strAddToGroup.empty() && _stricmp(name.c_str(), strAddToGroup.c_str()) != NULL)
    {   
        return configure(strAddToGroup.c_str(), "", pDefaultPaintThumbnail, pPaintThumbnailDefaults);
    }

    bool bDifferenceToDefault = false;

	 //Create Dummy Default
    bool bDefaultPaintThumbnailCreated = false;
    if ( pDefaultPaintThumbnail == NULL )
    {
        bDefaultPaintThumbnailCreated = true;
        if ( pPaintThumbnailDefaults == NULL )
            pDefaultPaintThumbnail = new PaintThumbnail();
        else
            pDefaultPaintThumbnail = new PaintThumbnail(pPaintThumbnailDefaults);
    }

	//-----------------------------------------------------------------
    m_strThumbnail_X = iGetx2RCString(name, (prefix + "ThumbnailX"), pDefaultPaintThumbnail->m_strThumbnail_X);
    if ( _stricmp(m_strThumbnail_X.c_str(), "0") != NULL )
        m_bSpecialSetup = true;
    if ( _stricmp(m_strThumbnail_X.c_str(), pDefaultPaintThumbnail->m_strThumbnail_X.c_str()) != NULL )
        bDifferenceToDefault = true;
    
    m_strThumbnail_Y = iGetx2RCString(name, (prefix + "ThumbnailY"), pDefaultPaintThumbnail->m_strThumbnail_Y);
    if ( _stricmp(m_strThumbnail_Y.c_str(), "0") != NULL )
        m_bSpecialSetup = true;
    if ( _stricmp(m_strThumbnail_Y.c_str(), pDefaultPaintThumbnail->m_strThumbnail_Y.c_str()) != NULL )
        bDifferenceToDefault = true;

    m_strThumbnail_Width = iGetx2RCString(name, (prefix + "ThumbnailWidth"), pDefaultPaintThumbnail->m_strThumbnail_Width);
    if ( _stricmp(m_strThumbnail_Width.c_str(), "100%") != NULL )
        m_bSpecialSetup = true;
    if ( _stricmp(m_strThumbnail_Width.c_str(), pDefaultPaintThumbnail->m_strThumbnail_Width.c_str()) != NULL )
        bDifferenceToDefault = true;
    
    m_strThumbnail_Height = iGetx2RCString(name, (prefix + "ThumbnailHeight"), pDefaultPaintThumbnail->m_strThumbnail_Height);
    if ( _stricmp(m_strThumbnail_Height.c_str(), "100%") != NULL )
        m_bSpecialSetup = true;
    if ( _stricmp(m_strThumbnail_Height.c_str(), pDefaultPaintThumbnail->m_strThumbnail_Height.c_str()) != NULL )
        bDifferenceToDefault = true;
    //-----------------------------------------------------------------

	m_iAlphaTransparency = iGetx2RCInt(name, (prefix + "ThumbnailAlphaTransparency"), pDefaultPaintThumbnail->m_iAlphaTransparency, 0, 255);
    if ( m_iAlphaTransparency != pDefaultPaintThumbnail->m_iAlphaTransparency )
        bDifferenceToDefault = true;

	if ( bDefaultPaintThumbnailCreated )
    {
        delete pDefaultPaintThumbnail;
        pDefaultPaintThumbnail = NULL;
    }
}

void PaintThumbnail::apply(HWND desthWnd, HWND sourcehWnd, int x, int y, int width, int height)
{
	if ( desthWnd == NULL || sourcehWnd == NULL )
			return;

	int iSpecialPosX = 0;
    int iSpecialPosY = 0;

    int iSpecialWidth = width;
    int iSpecialHeight = height;

    if ( m_bSpecialSetup )
    {
        iSpecialPosX = xParseCoord(m_strThumbnail_X.c_str(), 0, width, false);
        iSpecialPosY = xParseCoord(m_strThumbnail_Y.c_str(), 0, height, true);

        iSpecialWidth = xParseDimen(m_strThumbnail_Width.c_str(), 0, width, false);
        iSpecialHeight = xParseDimen(m_strThumbnail_Height.c_str(), 0, height, true);
    }

	//Register Thumbnail
	if ( m_bThumbnailRegistered /*&& thumbnail != NULL*/ )
	{
		if ( S_OK != zDwmUnregisterThumbnail(thumbnail) )
			MessageBox(NULL, "DwmUnregisterThumbnail failed!", "Info", MB_OK);
		m_bThumbnailRegistered = false;
	}

	HRESULT hr = zDwmRegisterThumbnail(desthWnd, &sourcehWnd, &thumbnail);
	if (SUCCEEDED(hr))
	{
		m_bThumbnailRegistered = true;

		SIZE sourceSize = {0,0};
		if ( S_OK != zDwmQueryThumbnailSourceSize(thumbnail, &sourceSize) )
			MessageBox(NULL, "DwmQueryThumbnailSourceSize failed!", "Info", MB_OK);
			
		int ratio = sourceSize.cx/sourceSize.cy;

		if ( ratio >= 1 )
		{
			iSpecialPosY = 	iSpecialPosY + ((iSpecialHeight-iSpecialWidth/ratio)/2);
			iSpecialHeight = iSpecialWidth/ratio;
		}
		else
		{
			iSpecialPosX = 	iSpecialPosX + ((iSpecialWidth-iSpecialHeight/ratio)/2);
			iSpecialWidth = iSpecialHeight/ratio;
		}

		//display thumbnail using DwmUpdateThumbnailProperties
		//destination rectangle size
		RECT dest;
		dest.left = x+iSpecialPosX;
		dest.top = y+iSpecialPosY;
		dest.right = dest.left+iSpecialWidth;
		dest.bottom = dest.top+iSpecialHeight;

		//Set thumbnail properties for use
		DWM_THUMBNAIL_PROPERTIES dskThumbProps;
		dskThumbProps.dwFlags = DWM_TNP_RECTDESTINATION | DWM_TNP_VISIBLE | DWM_TNP_SOURCECLIENTAREAONLY;
		//use window frame and client area
		dskThumbProps.fSourceClientAreaOnly = FALSE;
		dskThumbProps.fVisible = TRUE;
		dskThumbProps.opacity = m_iAlphaTransparency;
		dskThumbProps.rcDestination = dest;

		//display the thumbnail
		hr = zDwmUpdateThumbnailProperties(thumbnail, &dskThumbProps);
		if (SUCCEEDED(hr))
		{
			//do more things
		}
		else
		{
			MessageBox(NULL, "DwmUpdateThumbnailProperties failed!", "Info", MB_OK);
		}
	}
	else if ( hr == E_INVALIDARG )
	{
		MessageBox(NULL, "DwmRegisterThumbnail failed due to HWND top-level problems!", "Info", MB_OK);
	}
	else
	{
		MessageBox(NULL, "DwmRegisterThumbnail failed due to unknown reasons!", "Info", MB_OK);
	}
}
